/*
 * Copyright (c) 2022 New Vector Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package im.vector.app.features.voicebroadcast.recording

import android.content.Context
import android.media.MediaRecorder
import android.os.Build
import androidx.annotation.RequiresApi
import im.vector.app.core.di.ActiveSessionHolder
import im.vector.app.features.session.coroutineScope
import im.vector.app.features.voice.AbstractVoiceRecorderQ
import im.vector.app.features.voicebroadcast.model.VoiceBroadcast
import im.vector.app.features.voicebroadcast.model.VoiceBroadcastEvent
import im.vector.app.features.voicebroadcast.model.VoiceBroadcastState
import im.vector.app.features.voicebroadcast.usecase.GetMostRecentVoiceBroadcastStateEventUseCase
import im.vector.lib.core.utils.timer.CountUpTimer
import kotlinx.coroutines.Job
import kotlinx.coroutines.flow.launchIn
import kotlinx.coroutines.flow.onEach
import org.matrix.android.sdk.api.extensions.tryOrNull
import org.matrix.android.sdk.api.session.content.ContentAttachmentData
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.TimeUnit

@RequiresApi(Build.VERSION_CODES.Q)
class VoiceBroadcastRecorderQ(
        context: Context,
        private val sessionHolder: ActiveSessionHolder,
        private val getVoiceBroadcastEventUseCase: GetMostRecentVoiceBroadcastStateEventUseCase
) : AbstractVoiceRecorderQ(context), VoiceBroadcastRecorder {

    private val session get() = sessionHolder.getActiveSession()
    private val sessionScope get() = session.coroutineScope

    private var voiceBroadcastStateObserver: Job? = null

    private var maxFileSize = 0L // zero or negative for no limit
    private var currentVoiceBroadcast: VoiceBroadcast? = null
    private var currentMaxLength: Int = 0

    override var currentSequence = 0
    override var recordingState = VoiceBroadcastRecorder.State.Idle
        set(value) {
            field = value
            listeners.forEach { it.onStateUpdated(value) }
        }
    override var currentRemainingTime: Long? = null
        set(value) {
            field = value
            listeners.forEach { it.onRemainingTimeUpdated(value) }
        }

    private val recordingTicker = RecordingTicker()
    private val listeners = CopyOnWriteArrayList<VoiceBroadcastRecorder.Listener>()

    override val outputFormat = MediaRecorder.OutputFormat.MPEG_4
    override val audioEncoder = MediaRecorder.AudioEncoder.HE_AAC

    override val fileNameExt: String = "mp4"

    override fun initializeRecord(roomId: String, attachmentData: ContentAttachmentData?) {
        super.initializeRecord(roomId, attachmentData)
        mediaRecorder?.setMaxFileSize(maxFileSize)
        mediaRecorder?.setOnInfoListener { _, what, _ ->
            when (what) {
                MediaRecorder.MEDIA_RECORDER_INFO_MAX_FILESIZE_APPROACHING -> onMaxFileSizeApproaching(roomId)
                MediaRecorder.MEDIA_RECORDER_INFO_NEXT_OUTPUT_FILE_STARTED -> onNextOutputFileStarted()
                else -> Unit // Nothing to do
            }
        }
    }

    override fun startRecordVoiceBroadcast(voiceBroadcast: VoiceBroadcast, chunkLength: Int, maxLength: Int) {
        // Stop recording previous voice broadcast if any
        if (recordingState != VoiceBroadcastRecorder.State.Idle) stopRecord()

        currentVoiceBroadcast = voiceBroadcast
        maxFileSize = (chunkLength * audioEncodingBitRate / 8).toLong()
        currentMaxLength = maxLength
        currentSequence = 1

        observeVoiceBroadcastStateEvent(voiceBroadcast)
    }

    override fun pauseRecord() {
        if (recordingState != VoiceBroadcastRecorder.State.Recording) return
        tryOrNull { mediaRecorder?.stop() }
        mediaRecorder?.reset()
        recordingState = VoiceBroadcastRecorder.State.Paused
        recordingTicker.pause()
        notifyOutputFileCreated()
    }

    override fun resumeRecord() {
        if (recordingState != VoiceBroadcastRecorder.State.Paused) return
        currentSequence++
        currentVoiceBroadcast?.let { startRecord(it.roomId) }
        recordingState = VoiceBroadcastRecorder.State.Recording
        recordingTicker.resume()
    }

    override fun stopRecord() {
        super.stopRecord()

        // Stop recording
        recordingState = VoiceBroadcastRecorder.State.Idle
        recordingTicker.stop()
        notifyOutputFileCreated()

        // Remove listeners
        listeners.clear()

        // Do not observe anymore voice broadcast changes
        voiceBroadcastStateObserver?.cancel()
        voiceBroadcastStateObserver = null

        // Reset data
        currentSequence = 0
        currentMaxLength = 0
        currentRemainingTime = null
        currentVoiceBroadcast = null
    }

    override fun release() {
        mediaRecorder?.setOnInfoListener(null)
        super.release()
    }

    override fun addListener(listener: VoiceBroadcastRecorder.Listener) {
        listeners.add(listener)
        listener.onStateUpdated(recordingState)
        listener.onRemainingTimeUpdated(currentRemainingTime)
    }

    override fun removeListener(listener: VoiceBroadcastRecorder.Listener) {
        listeners.remove(listener)
    }

    private fun observeVoiceBroadcastStateEvent(voiceBroadcast: VoiceBroadcast) {
        voiceBroadcastStateObserver = getVoiceBroadcastEventUseCase.execute(voiceBroadcast)
                .onEach { onVoiceBroadcastStateEventUpdated(voiceBroadcast, it.getOrNull()) }
                .launchIn(sessionScope)
    }

    private fun onVoiceBroadcastStateEventUpdated(voiceBroadcast: VoiceBroadcast, event: VoiceBroadcastEvent?) {
        when (event?.content?.voiceBroadcastState) {
            VoiceBroadcastState.STARTED -> {
                startRecord(voiceBroadcast.roomId)
                recordingState = VoiceBroadcastRecorder.State.Recording
                recordingTicker.start()
            }
            VoiceBroadcastState.PAUSED -> pauseRecord()
            VoiceBroadcastState.RESUMED -> resumeRecord()
            VoiceBroadcastState.STOPPED,
            null -> stopRecord()
        }
    }

    private fun onMaxFileSizeApproaching(roomId: String) {
        setNextOutputFile(roomId)
    }

    private fun onNextOutputFileStarted() {
        notifyOutputFileCreated()
        currentSequence++
    }

    private fun notifyOutputFileCreated() {
        outputFile?.let { file ->
            listeners.forEach { it.onVoiceMessageCreated(file, currentSequence) }
            outputFile = nextOutputFile
            nextOutputFile = null
        }
    }

    private fun onElapsedTimeUpdated(elapsedTimeMillis: Long) {
        currentRemainingTime = if (currentMaxLength > 0 && recordingState != VoiceBroadcastRecorder.State.Idle) {
            val currentMaxLengthMillis = TimeUnit.SECONDS.toMillis(currentMaxLength.toLong())
            val remainingTimeMillis = currentMaxLengthMillis - elapsedTimeMillis
            TimeUnit.MILLISECONDS.toSeconds(remainingTimeMillis)
        } else {
            null
        }
    }

    private inner class RecordingTicker(
            private var recordingTicker: CountUpTimer? = null,
    ) {
        fun start() {
            recordingTicker?.stop()
            recordingTicker = CountUpTimer().apply {
                tickListener = CountUpTimer.TickListener { onTick(elapsedTime()) }
                resume()
                onTick(elapsedTime())
            }
        }

        fun pause() {
            recordingTicker?.apply {
                pause()
                onTick(elapsedTime())
            }
        }

        fun resume() {
            recordingTicker?.apply {
                resume()
                onTick(elapsedTime())
            }
        }

        fun stop() {
            recordingTicker?.apply {
                stop()
                onTick(elapsedTime())
                recordingTicker = null
            }
        }

        private fun onTick(elapsedTimeMillis: Long) {
            onElapsedTimeUpdated(elapsedTimeMillis)
        }
    }
}
